"""
Tests for pcmci.py, including tests for run_pc_stable, run_mci, and run_pcmci.
"""
from __future__ import print_function
from collections import defaultdict
from distutils.version import LooseVersion
import numpy as np
import pytest
from scipy.special import comb

from tigramite.pcmci import PCMCI
from tigramite.independence_tests.parcorr import ParCorr
import tigramite.data_processing as pp
from tigramite.toymodels import structural_causal_processes as toys

from test_pcmci_calculations import a_chain

# Pylint settings
# pylint: disable=redefined-outer-name

# Define the verbosity at the global scope
VERBOSITY = 1

# CONVENIENCE FUNCTIONS ########################################################
def _get_parent_graph(parents_neighbors_coeffs, exclude=None):
    """
    Iterates through the input parent-neighghbour coefficient dictionary to
    return only parent relations (i.e. where tau != 0)
    """
    graph = defaultdict(list)
    for j, i, tau, _ in toys._iter_coeffs(parents_neighbors_coeffs):
        if tau != 0 and (i, tau) != exclude:
            graph[j].append((i, tau))
    return dict(graph)

def _select_links(link_ids, true_parents):
    """
    Select links given by  from the true parents dictionary
    """
    if link_ids is None:
        return None
    return {par : {true_parents[par][link]:'-->'} for par in true_parents \
                                            for link in link_ids}

# TEST LINK GENERATION #########################################################
N_NODES = 4
@pytest.fixture(params=[
    # Generate a test data sample
    # Parameterize the sample by setting the autocorrelation value, coefficient
    # value, total time length, and random seed to different numbers
    # links_coeffs,     time, seed_val
    (a_chain(0.5, 0.6, length=N_NODES), 1000, 42)])
def a_sample(request):
    # Set the parameters
    links_coeffs, time, seed_val = request.param
    # Set the random seed
    np.random.seed(seed_val)
    # Generate the data
    data, _ = toys.var_process(links_coeffs, T=time)
    # Get the true parents
    true_parents = _get_parent_graph(links_coeffs)
    return pp.DataFrame(data), true_parents

# PCMCI CONSTRUCTION ###########################################################
@pytest.fixture()
    # Fixture to build and return a parameterized PCMCI.  Different selected
    # variables can be defined here.
def a_pcmci(a_sample, request):
    # Unpack the test data and true parent graph
    dataframe, true_parents = a_sample
    # Build the PCMCI instance
    pcmci = PCMCI(dataframe=dataframe,
                  cond_ind_test=ParCorr(verbosity=VERBOSITY),
                  verbosity=VERBOSITY)
    # Return the constructed PCMCI, expected results, and common parameters
    return pcmci, true_parents

# TEST VARIABLE SELECTION ######################################################
def get_select_vars(some_vars):
    """
    Yield some subsets of the input variables
    """
    for subset in [some_vars[0::2], some_vars[1::2]]:
        yield subset


# TEST LINK SELECTION ##########################################################
TAU_MIN = 1
TAU_MAX = 3
def get_expected_links(a_range, b_range, t_range):
    """
    Helper function to generate the expected links
    """
    return {a_var : {(b_var, -lag):'-?>' for b_var in b_range for lag in t_range}
            for a_var in a_range}

def test_select_links_default(a_pcmci):
    """
    Check that _set_sel_links returns all links by default.
    """
    # Unpack the pcmci instance
    pcmci, _ = a_pcmci
    # Test the default selected variables are correct
    sel_links = pcmci._set_link_assumptions(None, TAU_MIN, TAU_MAX)
    err_msg = "Default option for _set_link_assumptions should return all possible "+\
              "combinations"
    good_links = get_expected_links(range(pcmci.N),
                                    range(pcmci.N),
                                    range(TAU_MIN, TAU_MAX + 1))
    np.testing.assert_equal(sel_links, good_links, err_msg=err_msg)

# TEST ITERATORS ###############################################################
@pytest.fixture(params=[
    # Store some parameters for setting maximum conditions
    #max_cond_dim, tau_min, tau_max
    (None,         1,       3),
    (10,           1,       3),
    (1,            2,       4)])
def a_iter_cond_param(request):
    return request.param

def test_condition_iterator(a_pcmci, a_iter_cond_param):
    """
    Test if the correct conditions are returned from the iterator.  Checks that:
        * Parent node never included in yielded conditions
        * Yielded conditions of the same dimension are unique
        * The expected total number of conditions are given
    """
    # Unpack the pcmci algorithm
    pcmci, _ = a_pcmci
    # Unpack the iterator conditions parameters
    max_cond_dim, tau_min, tau_max = a_iter_cond_param
    # Get all possible links
    sel_links = pcmci._set_link_assumptions(None, tau_min, tau_max)
    # Loop over all possible condition dimentions
    max_cond_dim = pcmci._set_max_condition_dim(max_cond_dim, tau_min, tau_max)
    # Loop over all nodes
    for j in range(pcmci.N):
        # Initialise the list of conditions for this node
        total_n_conds = 0
        expected_total_n_conds = 0
        # Parents for this node
        parents = sel_links[j]
        # Loop over all possible dimentionality of conditions
        for cond_dim in range(max_cond_dim+1):
            # Skip the case where (cond_dim == num_parents + 1), as is done in 
            # the main loop of the code
            if cond_dim > len(parents) - 1:
                continue
            # Iterate through all possible pairs (that have not converged yet)
            for par in parents:
                # Number of conditions for this dimension
                expected_n_conds = comb(len(parents) - 1, cond_dim, exact=True)
                expected_total_n_conds += expected_n_conds
                # Create an array for these conditions
                par_conds = list()
                # Iterate through all possible combinations
                for cond in pcmci._iter_conditions(par, cond_dim, parents):
                    # Ensure the condition does not contain the parent
                    assert par not in cond, "Parent should not be in condition"
                    # Keep this condition in list of conditions
                    par_conds.append([par] + cond)
                    assert len(cond) == cond_dim,\
                        "Number of conditions should match requested dimension"
                # Check that at least one condition is found
                assert len(par_conds) != 0,\
                    "Expected number of conditions for condition "+\
                    " dimension {}".format(cond_dim)+\
                    " and total n_nodes {}".format(len(parents)-1)+\
                    " is zero!"
                # Check that this iterations has the expected number of
                # conditions
                assert len(par_conds) == expected_n_conds,\
                    "Expected number of conditions for condition "+\
                    " dimension {}".format(cond_dim)+\
                    " and total n_nodes {}".format(len(parents)-1)+\
                    " does not match"
                # Record the total number found
                total_n_conds += len(par_conds)
                par_conds = np.array(par_conds)
                # Check that all condition, parent pairs are unique
                # TODO fix this for 3.4 support
                if LooseVersion(np.version.version) >= LooseVersion("1.13"):
                    assert np.unique(par_conds, axis=0).shape == par_conds.shape
        # Check that the right number of conditions are found
        assert total_n_conds == expected_total_n_conds, \
            "Expected number of combinations found"

@pytest.fixture(params=[
    # Store some parameters for iterating over conditions during the MCI
    # algorithm
    #max_cond_dim_X, max_cond_dim_Y, tau_min, tau_max
    (None,           None,           0,       3), # Default
    (1,              None,           0,       3), # Maximum conds on X
    (None,           1,              0,       3), # Maximum conds on Y
    (None,           None,           1,       3), # Non-trivial tau-min
    (None,           None,           0,       1)])# Non-trivial tau-max
def a_iter_indep_cond_param(request):
    return request.param

def test_iter_indep_conds(a_pcmci, a_iter_indep_cond_param):
    """
    Checks that the iteration for the MCI-like part of the algorithm works.
    This tests that:
        * The correct number of link-parent pairs are tested
        * The parent node is not in the condition
        * The number of conditions is less than the implied maximum
    """
    # Unpack the pcmci algorithm
    pcmci, parents = a_pcmci
    # Unpack the parameters
    max_cx, max_cy, tau_min, tau_max = a_iter_indep_cond_param
    # Set the link assumptions
    _int_set_link_assumptions = pcmci._set_link_assumptions(None, tau_min, tau_max)
    # Set the maximum condition dimension for Y and Z
    max_conds_py = pcmci._set_max_condition_dim(max_cx, tau_min, tau_max)
    max_conds_px = pcmci._set_max_condition_dim(max_cy, tau_min, tau_max)
    # Get the parents that will be checked
    _int_parents = pcmci._get_int_parents(parents)
    # Get the conditions as implied by the input arguments
    n_links = 0
    # Calculated the expected number of links to test.
    expect_links = 0
    # Since a link is never checked as a parent of itself, remove all
    # (i, tau) == (j, 0) entries from expected testing
    for j, node_list in _int_set_link_assumptions.items():
        for i, tau in node_list:
            if not (i == j and  tau == 0):
                expect_links += 1
    ## TODO try to check content of Z to ensure it contains parent of y, parents 
    ## of x lagged.
    # Iterate over all the returned conditions
    for j, i, tau, Z in pcmci._iter_indep_conds(_int_parents,
                                                _int_set_link_assumptions,
                                                max_conds_py,
                                                max_conds_px):
        # Incriment the link count
        n_links += 1
        # Ensure the parent link is not in the conditions
        assert (i, tau) not in Z,\
            "Parent node must not be in the conditions"
        # Check that the conditions length are shorter than the maximum
        # condition length for X and Y
        total_max_conds = max_conds_px + max_conds_py
        assert len(Z) <= total_max_conds,\
            "The number of conditions must be less than the sum of the"+\
            " maximum number of conditions on X and Y"
    # Get the total number of tested links
    assert expect_links == n_links,\
        "Recoved the expected number of links to test"

# TEST UTILITY FUNCTIONS #######################################################
def test_sort_parents(a_pcmci):
    """
    Test that the function that sorts parents returns the sorted values as
    desired
    """
    # Unpack the pcmci instance
    pcmci, _ = a_pcmci
    # Create some parents to sort
    orig_parents = []
    n_parents = 10
    for i in range(n_parents):
        orig_parents.append((i, i))
    # Put them into a dictionary
    parent_vals = {}
    sign = 1
    for val, par in enumerate(orig_parents):
        # Alternate the sign to ensure sorting is done on abolute value
        sign *= -1
        # Use greater values of the test metric for greater values of the parent
        # indexes
        parent_vals[par] = val * sign
    # Sort them
    sorted_parents = pcmci._sort_parents(parent_vals)
    # Ensure the sorting works as expected
    assert sorted_parents == orig_parents[::-1],\
        "Parents must be sorted by abolute value of the test metric"

@pytest.fixture(params=[
    # Store some parameters for iterating over conditions during the MCI
    # algorithm
    #tau_min, tau_max, should_pass
    (0,       3,       True),  # 0 = tau_min < tau_max, should pass
    (2,       3,       True),  # 0 < tau_min < tau_max, should pass
    (3,       3,       True),  # 0 < tau_min = tau_max, should pass
    (3,       2,       False), # 0 < tau_max < tau_min, should fail
    (-1,      3,       False), # tau_min < 0 < tau_min, should fail
    (0,      -2,       False), # tau_max < 0 < tau_min, should fail
    (-1,     -2,       False)])# tau_min < tau_max < 0, should fail
def a_tau_values(request):
    return request.param

def test_check_tau_limits(a_pcmci, a_tau_values):
    """
    Test the tau limit checker fails correctly on:
        * Negative values
        * Values where tau_min > tau_max
    And passes otherwise
    """
    # Unpack the pcmci instance
    pcmci, _ = a_pcmci
    # Unpack the tau parameters
    tau_min, tau_max, should_pass = a_tau_values
    err_msg = "\ntau_min : {}\ntau_max : {}".format(tau_min, tau_max)
    # If it should pass, try it
    if should_pass:
        try:
            pcmci._check_tau_limits(tau_min, tau_max)
        # Ensure no exception is raised
        except:
            pytest.fail("Good tau limits failed incorrectly:"+err_msg)
    # If it should fail, make sure it does
    else:
        with pytest.raises(ValueError):
            pcmci._check_tau_limits(tau_min, tau_max)
            pytest.fail("Bad tau limits should fail"+err_msg)

@pytest.fixture(params=[
    # Store some parameters for correcting the pvalues
    #fdr_method, excl,  slice_n,        slice_t,     message
    ('fdr_bh',   False, range(N_NODES), 0,           "default"),
    ('fdr_bh',   True,  slice(None),    0,           "exclude contemporaneous"),
    ('none',     True,  slice(None),    slice(None), "none")])
def a_correct_pvals_params(request):
    return request.param

def test_corrected_pvalues(a_pcmci, a_correct_pvals_params):
    """
    Test the wrapper functions of the corrected pvalue function.  This means:
        * check 'none' is a valid mode that does nothing to the input
        * check that exclude_contemporaneous works
        * check that autocorrelation elements are ignored
    """
    # Unpack the pcmci instance
    pcmci, _ = a_pcmci
    # Unpack the parameters
    fdr_method, excl, slice_n, slice_t, message = a_correct_pvals_params
    # Create the p-values
    pvals = np.linspace(0, 1, num=N_NODES*N_NODES*(TAU_MAX+1))
    pvals = pvals.reshape(N_NODES, N_NODES, TAU_MAX+1)
    # Create the corrected p-values
    qvals = pcmci.get_corrected_pvalues(pvals,
                                        fdr_method=fdr_method,
                                        exclude_contemporaneous=excl)
    err_msg = "get_corrected_pvalues failed on "+message+" mode"
    np.testing.assert_allclose(pvals[slice_n, slice_n, slice_t],
                               qvals[slice_n, slice_n, slice_t],
                               rtol=1e-10,
                               atol=1e-10,
                               verbose=True,
                               err_msg=err_msg)

def test_sig_parents(a_pcmci):
    """
    Test that the correct significant parents are returned.
    """
    # Unpack the pcmci instance
    pcmci, _ = a_pcmci
    # Define the dimensionality
    dim = N_NODES
    # Build a p_matrix for 10 x 10 x 10
    p_matrix = np.arange(dim*dim*dim).reshape(dim, dim, dim) + 1
    # Build a val matrix as the negative version of this matrix
    val_matrix = -p_matrix
    # Define the alpha value
    alpha = dim*dim*dim/2
    # Get the significant parents
    graph = pcmci.get_graph_from_pmatrix(p_matrix=p_matrix, tau_min=0, tau_max=dim-1,
                                                   alpha_level=alpha)
    # # Ensure the link matrix has the correct sum
    # link_matrix = graph != ""
    # num_links = np.count_nonzero(link_matrix)
    # assert num_links == alpha,\
    #     "The correct number of significant parents are found in the returned"+\
    #     " link matrix"
    # # Ensure all the parents are in the second half of the returned p_matrix
    # num_links = np.count_nonzero(link_matrix[:5, :, :])
    # assert num_links == alpha,\
    #     "The correct links from significant parents are found in the returned"+\
    #     " link matrix"
    # Ensure the correct number of links are returned in the dictionary of
    # parents
    parents_dict = pcmci.return_parents_dict(graph=graph, val_matrix=val_matrix)
    all_links = [lnk for links in parents_dict.values() for lnk in links]
    assert len(all_links) == (dim*dim*(dim - 1))/2.,\
            "The correct number of links are returned in the dictionary of"+\
            " parents"
    # Ensure the correct links are returned:
    # Expect j to cycle through [0, dim]
    expect_j = set(range(dim))
    # Expect i to cycle through [0, dim/2]
    expect_i = set(range((int)(dim/2)))
    # Expect tau to cycle through [-3, -1]
    expect_t = set(range(-dim + 1, 0))
    for key, links in parents_dict.items():
        # Ensure the returned keys are correct
        assert key in expect_j, "Incorrect node/key found in returned parent"+\
                " dictionary"
        for (i, tau) in links:
            # Ensure the returned node values are correct
            assert i in expect_i, "Incorrect parent found in returned parent"+\
                    " dictionary"
            # Ensure the returned tau values are correct
            assert tau in expect_t, "Incorrect tau found in returned parent"+\
                    " dictionary"



"""Tigramite toymodels."""

# Author: Jakob Runge <jakob@jakob-runge.com>
#
# License: GNU General Public License v3.0
from __future__ import print_function
import sys
import warnings
import copy
import math
import numpy as np

from tigramite.toymodels import structural_causal_processes as toys



def generate_linear_model_from_data(dataframe, parents, tau_max, realizations=100, 
                generate_noise_from='covariance',  
                T_data = None,
                model_params=None,
                data_transform=None,
                mask_type='y',
                boot_blocklength=1,
                seed=42,
                verbosity=0):
    """
    Fits a (contemporaneous and lagged) linear SCM to data, computes
    residuals, and then generates surrogate realizations with noise drawn
    with replacement from residuals or from a multivariate normal.
    
    Parameters
    ----------
    generate_noise_from : {'covariance', 'residuals'}
        Whether to generate the noise from a gaussian with same mean and covariance
        as residuals, or by drawing (with replacement) from the residuals.
    boot_blocklength : int, or in {'cube_root', 'from_autocorrelation'}
        Block length for block-bootstrap, which only applies to
        generate_noise_from='residuals'. If 'from_autocorrelation', the block
        length is determined from the decay of the autocovariance and
        if 'cube_root' it is the cube root of the time series length.
    seed : int, optional(default = None)
        Seed for RandomState (default_rng)
    """

    from tigramite.models import Models
    from sklearn.linear_model import LinearRegression

    assert dataframe.analysis_mode == 'single'

    random_state = np.random.default_rng(seed)

    def lin_f(x): return x

    if model_params is None:
        model_params = {}

    N = dataframe.N
    T = dataframe.T[0]
    if T_data is None:
        T_data = T

    ## Estimate noise covariance matrix of residuals
    model = Models(dataframe=dataframe,
                     model=LinearRegression(**model_params),
                     data_transform=data_transform,
                     mask_type='y',
                     verbosity=0)

    model.fit_full_model(all_parents=parents, tau_max=tau_max, return_data=True)

    links_coeffs = {}
    for j in range(N):
        links_coeffs[j] = []
        if len(parents[j]) > 0:
            for ipar, par in enumerate(parents[j]):
                links_coeffs[j].append(((par[0], int(par[1])), model.fit_results[j]['model'].coef_[ipar], lin_f))
                if verbosity > 0:
                    print(j, ((par[0], int(par[1])), np.round(model.fit_results[j]['model'].coef_[ipar], 2),) )

    pred = model.predict_full_model(
                new_data=None,
                pred_params=None,
                cut_off='max_lag_or_tau_max')

    # Computes cov and mean, but internally also the residuals needed here
    cov, mean = model.get_residuals_cov_mean(new_data=None,
                pred_params=None,)

    if generate_noise_from == 'covariance':
        # cov = np.cov(overlapping_residuals, rowvar=0)
        # mean = np.mean(overlapping_residuals, axis=0)   # residuals should have zero mean due to prediction including constant
        if verbosity > 0:
            print('covariance')
            print(np.round(cov, 2))
            print('mean')
            print(np.round(mean, 2))

    overlapping_residuals = model.residuals
    len_residuals = len(overlapping_residuals)

    ## Construct linear Gaussian structural causal model with this noise structure and generate many realizations with same sample size as data
    transient_fraction = 0.2
    size = T_data + int(math.floor(transient_fraction*T_data))
    # datasets = {}

    if generate_noise_from == 'residuals':
        if boot_blocklength == 'cube_root':
            boot_blocklength = max(1, int(T_data**(1/3)))
        elif boot_blocklength == 'from_autocorrelation':
            boot_blocklength = \
                get_block_length(overlapping_residuals.T, xyz=np.zeros(N), mode='confidence')
        elif type(boot_blocklength) is int and boot_blocklength > 0:
            pass
        else:
            raise ValueError("boot_blocklength must be integer > 0, 'cube_root', or 'from_autocorrelation'")

        # Determine the number of blocks total, rounding up for non-integer
        # amounts
        n_blks = int(math.ceil(float(size) / boot_blocklength))
        if n_blks < 10:
            raise ValueError("Only %d block(s) for block-sampling,"  %n_blks +
                             "choose smaller boot_blocklength!")

    for r in range(realizations):
        if generate_noise_from == 'covariance':
            noises = random_state.multivariate_normal(mean=mean, cov=cov, size=size)
        elif generate_noise_from == 'residuals':

            # Get the starting indices for the blocks
            blk_strt = random_state.choice(np.arange(len_residuals - boot_blocklength), 
                size=n_blks, replace=True)

            # Get the empty array of block resampled values
            boot_draw = np.zeros(n_blks*boot_blocklength, dtype='int')
            # Fill the array of block resamples
            for i in range(boot_blocklength):
                boot_draw[i::boot_blocklength] = np.arange(size)[blk_strt + i]
            # Cut to proper length
            draw = boot_draw[:size]
            # draw = np.random.randint(0, len(overlapping_residuals), size)    
            noises = overlapping_residuals[draw]

        else: raise ValueError("generate_noise_from has to be either 'covariance' or 'residuals'")

        dataset = toys.structural_causal_process(links=links_coeffs, noises=noises, T=T_data, 
                                                     transient_fraction=transient_fraction)[0]
        if np.any(np.isinf(dataset)):
            raise ValueError("Infinite data")

        yield dataset

    # return self   #datasets


def get_acf(series, max_lag=None):
    """Returns autocorrelation function.

    Parameters
    ----------
    series : 1D-array
        data series to compute autocorrelation from

    max_lag : int, optional (default: None)
        maximum lag for autocorrelation function. If None is passed, 10% of
        the data series length are used.

    Returns
    -------
    autocorr : array of shape (max_lag + 1,)
        Autocorrelation function.
    """
    # Set the default max lag
    if max_lag is None:
        max_lag = int(max(5, 0.1*len(series)))
    # Initialize the result
    autocorr = np.ones(max_lag + 1)
    # Iterate over possible lags
    for lag in range(1, max_lag + 1):
        # Set the values
        y1_vals = series[lag:]
        y2_vals = series[:len(series) - lag]
        # Calculate the autocorrelation
        autocorr[lag] = np.corrcoef(y1_vals, y2_vals)[0, 1]
    return autocorr

def get_block_length(array, xyz, mode):
    """Returns optimal block length for significance and confidence tests.

    Determine block length using approach in Mader (2013) [Eq. (6)] which
    improves the method of Pfeifer (2005) with non-overlapping blocks In
    case of multidimensional X, the max is used. Further details in [1]_.
    Two modes are available. For mode='significance', only the indices
    corresponding to X are shuffled in array. For mode='confidence' all
    variables are jointly shuffled. If the autocorrelation curve fit fails,
    a block length of 5% of T is used. The block length is limited to a
    maximum of 10% of T.

    Mader et al., Journal of Neuroscience Methods,
    Volume 219, Issue 2, 15 October 2013, Pages 285-291

    Parameters
    ----------
    array : array-like
        data array with X, Y, Z in rows and observations in columns

    xyz : array of ints
        XYZ identifier array of shape (dim,).

    mode : str
        Which mode to use.

    Returns
    -------
    block_len : int
        Optimal block length.
    """
    # Inject a dependency on siganal, optimize
    from scipy import signal, optimize
    # Get the shape of the array
    dim, T = array.shape
    # Initiailize the indices
    indices = range(dim)
    if mode == 'significance':
        indices = np.where(xyz == 0)[0]

    # Maximum lag for autocov estimation
    max_lag = int(0.1*T)
    # Define the function to optimize against
    def func(x_vals, a_const, decay):
        return a_const * decay**x_vals

    # Calculate the block length
    block_len = 1
    for i in indices:
        # Get decay rate of envelope of autocorrelation functions
        # via hilbert trafo
        autocov = get_acf(series=array[i], max_lag=max_lag)
        autocov[0] = 1.
        hilbert = np.abs(signal.hilbert(autocov))
        # Try to fit the curve
        try:
            popt, _ = optimize.curve_fit(
                f=func,
                xdata=np.arange(0, max_lag+1),
                ydata=hilbert,
            )
            phi = popt[1]
            # Formula of Mader (2013) assuming non-overlapping blocks
            l_opt = (4. * T * (phi / (1. - phi) + phi**2 / (1. - phi)**2)**2
                     / (1. + 2. * phi / (1. - phi))**2)**(1. / 3.)
            block_len = max(block_len, int(l_opt))

            # pylab.plot(np.arange(0, max_lag+1), hilbert)
            # pylab.plot(np.arange(0, max_lag+1), func(np.arange(0, max_lag+1), popt[0], popt[1]))
            # print("block_len ", block_len, int(l_opt))
            # pylab.show()
        except RuntimeError:
            warnings.warn("Error - curve_fit failed for estimating block_shuffle length, using"
                  " block_len = %d" % (int(.05 * T)))
            # block_len = max(int(.05 * T), block_len)

    # print("chosen ", block_len)
    # Limit block length to a maximum of 10% of T
    block_len = min(block_len, int(0.1 * T))
    # print("chosen ", block_len)

    return block_len


if __name__ == '__main__':


    import tigramite
    from tigramite import data_processing as pp

    lin_f = lambda x: x
    links_coeffs = {0: [((0, -1), 0.98, lin_f), ((0, -2), -0.7, lin_f)],
                    1: [((1, -1), 0.9, lin_f), ((0, -1), 0.3, lin_f)],
                    2: [((2, -1), 0.9, lin_f), ((0, -2), -0.5, lin_f)],
                    3: [((3, -1), 0.9, lin_f)], #, ((4, -1), 0.4, lin_f)],
                    4: [((4, -1), 0.9, lin_f), ((3, 0), 0.5, lin_f)], #, ((3, -1), 0.3, lin_f)],
                    }
    T = 50     # time series length
    # Make some noise with different variance, alternatively just noises=None
    noises = None  # np.array([(1. + 0.2*float(j))*np.random.randn((T + int(math.floor(0.2*T)))) 
                       # for j in range(len(links_coeffs))]).T

    data, _ = toys.structural_causal_process(links_coeffs, T=T, noises=noises)
    T, N = data.shape

    # For generality, we include some masking
    mask = np.zeros(data.shape, dtype='int')
    mask[:int(T/2),0] = True
    # mask[int(T/2)+30:,1] = True
    # Create some missing samples at different time points
    data[11,0] = 9999.
    data[22,2] = 9999.
    data[33,3] = 9999.

    tau_max = 4
    # Initialize dataframe object, specify time axis and variable names
    var_names = [r'$X^0$', r'$X^1$', r'$X^2$', r'$X^3$', r'$X^4$']
    dataframe = pp.DataFrame(data, 
                             mask=mask,
                             missing_flag = 9999.,
                             datatime = {0:np.arange(len(data))}, 
                             var_names=var_names)
    parents = {}
    for j in links_coeffs:
        parents[j] = []
        for par in links_coeffs[j]:
            parents[j].append(par[0])
    print(parents)
    datasets = list(generate_linear_model_from_data(dataframe, parents=parents, 
                tau_max=tau_max, realizations=100, 
                generate_noise_from='residuals',
                boot_blocklength=3, #'from_autocorrelation',
                verbosity=0))
    print(datasets[0].shape)



"""Tigramite causal discovery for time series."""

# Author: Jakob Runge <jakob@jakob-runge.com>
#
# License: GNU General Public License v3.0

from __future__ import print_function
import json, warnings, os, pathlib
import numpy as np
import gc
import dcor
import torch
import gpytorch
from .LBFGS import FullBatchLBFGS
from .independence_tests_base import CondIndTest

class GaussProcRegTorch():
    r"""Gaussian processes abstract base class.

    GP is estimated with gpytorch. Note that the kernel's hyperparameters are
    optimized during fitting.

    When the null distribution is not analytically available, but can be
    precomputed with the function generate_and_save_nulldists(...) which saves
    a \*.npz file containing the null distribution for different sample sizes.
    This file can then be supplied as null_dist_filename.

    Assumes one-dimensional X, Y. But can be combined with PairwiseMultCI to
    obtain a test for multivariate X, Y.

    Parameters
    ----------
    null_samples : int
        Number of null samples to use

    cond_ind_test : CondIndTest
        Conditional independence test that this Gaussian Proccess Regressor will
        calculate the null distribution for.  This is used to grab the
        get_dependence_measure function.

    null_dist_filename : str, otional (default: None)
        Path to file containing null distribution.

    verbosity : int, optional (default: 0)
        Level of verbosity.
    """

    def __init__(self,
                 null_samples,
                 cond_ind_test,
                 null_dist_filename=None,
                 checkpoint_size=None,
                 verbosity=0):
        # Set the dependence measure function
        self.cond_ind_test = cond_ind_test
        # Set member variables
        self.verbosity = verbosity
        # Set the null distribution defaults
        self.null_samples = null_samples
        self.null_dists = {}
        self.null_dist_filename = null_dist_filename
        # Check if we are loading a null distrubtion from a cached file
        if self.null_dist_filename is not None:
            self.null_dists, self.null_samples = \
                self._load_nulldist(self.null_dist_filename)
        # Size for batching
        self.checkpoint_size = checkpoint_size

    def _load_nulldist(self, filename):
        r"""
        Load a precomputed null distribution from a \*.npz file.  This
        distribution can be calculated using generate_and_save_nulldists(...).

        Parameters
        ----------
        filename : strng
            Path to the \*.npz file

        Returns
        -------
        null_dists, null_samples : dict, int
            The null distirbution as a dictionary of distributions keyed by
            sample size, the number of null samples in total.
        """
        null_dist_file = np.load(filename)
        null_dists = dict(zip(null_dist_file['T'],
                              null_dist_file['exact_dist']))
        null_samples = len(null_dist_file['exact_dist'][0])
        return null_dists, null_samples

    def _generate_nulldist(self, df,
                           add_to_null_dists=True):
        """Generates null distribution for pairwise independence tests.

        Generates the null distribution for sample size df. Assumes pairwise
        samples transformed to uniform marginals. Uses get_dependence_measure
        available in class and generates self.sig_samples random samples. Adds
        the null distributions to self.null_dists.

        Parameters
        ----------
        df : int
            Degrees of freedom / sample size to generate null distribution for.
        add_to_null_dists : bool, optional (default: True)
            Whether to add the null dist to the dictionary of null dists or
            just return it.

        Returns
        -------
        null_dist : array of shape [df,]
            Only returned,if add_to_null_dists is False.
        """

        if self.verbosity > 0:
            print("Generating null distribution for df = %d. " % df)
            if add_to_null_dists:
                print("For faster computations, run function "
                      "generate_and_save_nulldists(...) to "
                      "precompute null distribution and load *.npz file with "
                      "argument null_dist_filename")

        xyz = np.array([0, 1])

        null_dist = np.zeros(self.null_samples)
        for i in range(self.null_samples):
            array = self.cond_ind_test.random_state.random((2, df))
            null_dist[i] = self.cond_ind_test.get_dependence_measure(
                array, xyz)

        null_dist.sort()
        if add_to_null_dists:
            self.null_dists[df] = null_dist
        return null_dist

    def _generate_and_save_nulldists(self, sample_sizes, null_dist_filename):
        """Generates and saves null distribution for pairwise independence
        tests.

        Generates the null distribution for different sample sizes. Calls
        generate_nulldist. Null dists are saved to disk as
        self.null_dist_filename.npz. Also adds the null distributions to
        self.null_dists.

        Parameters
        ----------
        sample_sizes : list
            List of sample sizes.

        null_dist_filename : str
            Name to save file containing null distributions.
        """

        self.null_dist_filename = null_dist_filename

        null_dists = np.zeros((len(sample_sizes), self.null_samples))

        for iT, T in enumerate(sample_sizes):
            null_dists[iT] = self._generate_nulldist(
                T, add_to_null_dists=False)
            self.null_dists[T] = null_dists[iT]

        np.savez("%s" % null_dist_filename,
                 exact_dist=null_dists,
                 T=np.array(sample_sizes))


    def _get_single_residuals(self, array, target_var,
                                    return_means=False,
                                    standardize=True,
                                    return_likelihood=False,
                                    training_iter=50,
                                    lr=0.1):
        """Returns residuals of Gaussian process regression.

        Performs a GP regression of the variable indexed by target_var on the
        conditions Z. Here array is assumed to contain X and Y as the first two
        rows with the remaining rows (if present) containing the conditions Z.
        Optionally returns the estimated mean and the likelihood.

        Parameters
        ----------
        array : array-like
            data array with X, Y, Z in rows and observations in columns

        target_var : {0, 1}
            Variable to regress out conditions from.

        standardize : bool, optional (default: True)
            Whether to standardize the array beforehand.

        return_means : bool, optional (default: False)
            Whether to return the estimated regression line.

        return_likelihood : bool, optional (default: False)
            Whether to return the log_marginal_likelihood of the fitted GP.

        training_iter : int, optional (default: 50)
            Number of training iterations.

        lr : float, optional (default: 0.1)
            Learning rate (default: 0.1).

        Returns
        -------
        resid [, mean, likelihood] : array-like
            The residual of the regression and optionally the estimated mean
            and/or the likelihood.
        """

        dim, T = array.shape

        if dim <= 2:
            if return_likelihood:
                return array[target_var, :], -np.inf
            return array[target_var, :]

        # Implement using PyTorch
        # Standardize
        if standardize:
            array -= array.mean(axis=1).reshape(dim, 1)
            std = array.std(axis=1)
            for i in range(dim):
                if std[i] != 0.:
                    array[i] /= std[i]
            if np.any(std == 0.) and self.verbosity > 0:
                warnings.warn("Possibly constant array!")
            # array /= array.std(axis=1).reshape(dim, 1)
            # if np.isnan(array).any():
            #     raise ValueError("Nans after standardizing, "
            #                      "possibly constant array!")

        target_series = array[target_var, :]
        z = array[2:].T.copy()
        if np.ndim(z) == 1:
            z = z.reshape(-1, 1)

        train_x = torch.tensor(z).float()
        train_y = torch.tensor(target_series).float()

        device_type = 'cuda' if torch.cuda.is_available() else 'cpu'
        output_device = torch.device(device_type)
        train_x, train_y = train_x.to(output_device), train_y.to(output_device)

        if device_type == 'cuda':
            # If GPU is available, use MultiGPU with Kernel Partitioning
            n_devices = torch.cuda.device_count()
            class mExactGPModel(gpytorch.models.ExactGP):
                def __init__(self, train_x, train_y, likelihood, n_devices):
                    super(mExactGPModel, self).__init__(train_x, train_y, likelihood)
                    self.mean_module = gpytorch.means.ConstantMean()
                    base_covar_module = gpytorch.kernels.ScaleKernel(gpytorch.kernels.RBFKernel())

                    self.covar_module = gpytorch.kernels.MultiDeviceKernel(
                        base_covar_module, device_ids=range(n_devices),
                        output_device=output_device
                    )

                def forward(self, x):
                    mean_x = self.mean_module(x)
                    covar_x = self.covar_module(x)
                    return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)

            def mtrain(train_x,
                      train_y,
                      n_devices,
                      output_device,
                      checkpoint_size,
                      preconditioner_size,
                      n_training_iter,
                      ):
                likelihood = gpytorch.likelihoods.GaussianLikelihood().to(output_device)
                model = mExactGPModel(train_x, train_y, likelihood, n_devices).to(output_device)
                model.train()
                likelihood.train()

                optimizer = FullBatchLBFGS(model.parameters(), lr=lr)
                # "Loss" for GPs - the marginal log likelihood
                mll = gpytorch.mlls.ExactMarginalLogLikelihood(likelihood, model)

                with gpytorch.beta_features.checkpoint_kernel(checkpoint_size), \
                        gpytorch.settings.max_preconditioner_size(preconditioner_size):

                    def closure():
                        optimizer.zero_grad()
                        output = model(train_x)
                        loss = -mll(output, train_y)
                        return loss

                    loss = closure()
                    loss.backward()

                    for i in range(n_training_iter):
                        options = {'closure': closure, 'current_loss': loss, 'max_ls': 10}
                        loss, _, _, _, _, _, _, fail = optimizer.step(options)

                        '''print('Iter %d/%d - Loss: %.3f   lengthscale: %.3f   noise: %.3f' % (
                            i + 1, n_training_iter, loss.item(),
                            model.covar_module.module.base_kernel.lengthscale.item(),
                            model.likelihood.noise.item()
                        ))'''

                        if fail:
                            # print('Convergence reached!')
                            break

                return model, likelihood, mll

            def find_best_gpu_setting(train_x,
                                      train_y,
                                      n_devices,
                                      output_device,
                                      preconditioner_size
                                      ):
                N = train_x.size(0)

                # Find the optimum partition/checkpoint size by decreasing in powers of 2
                # Start with no partitioning (size = 0)
                settings = [0] + [int(n) for n in np.ceil(N / 2 ** np.arange(1, np.floor(np.log2(N))))]

                for checkpoint_size in settings:
                    print('Number of devices: {} -- Kernel partition size: {}'.format(n_devices, checkpoint_size))
                    try:
                        # Try a full forward and backward pass with this setting to check memory usage
                        _, _, _ = mtrain(train_x, train_y,
                                     n_devices=n_devices, output_device=output_device,
                                     checkpoint_size=checkpoint_size,
                                     preconditioner_size=preconditioner_size, n_training_iter=1)

                        # when successful, break out of for-loop and jump to finally block
                        break
                    except RuntimeError as e:
                        pass
                    except AttributeError as e:
                        pass
                    finally:
                        # handle CUDA OOM error
                        gc.collect()
                        torch.cuda.empty_cache()
                return checkpoint_size

            # Set a large enough preconditioner size to reduce the number of CG iterations run
            preconditioner_size = 100
            if self.checkpoint_size is None:
                self.checkpoint_size = find_best_gpu_setting(train_x, train_y,
                                                        n_devices=n_devices,
                                                        output_device=output_device,
                                                        preconditioner_size=preconditioner_size)

            model, likelihood, mll = mtrain(train_x, train_y,
                                      n_devices=n_devices, output_device=output_device,
                                      checkpoint_size=self.checkpoint_size,
                                      preconditioner_size=100,
                                      n_training_iter=training_iter)

            # Get into evaluation (predictive posterior) mode
            model.eval()
            likelihood.eval()

            # Make predictions by feeding model through likelihood
            with torch.no_grad(), gpytorch.settings.fast_pred_var(), gpytorch.beta_features.checkpoint_kernel(1000):
                mean = model(train_x).loc.detach()
                loglik = mll(model(train_x), train_y)*T

            resid = (train_y - mean).detach().cpu().numpy()
            mean = mean.detach().cpu().numpy()

        else:
            # If only CPU is available, we will use the simplest form of GP model, exact inference
            class ExactGPModel(gpytorch.models.ExactGP):
                def __init__(self, train_x, train_y, likelihood):
                    super(ExactGPModel, self).__init__(
                        train_x, train_y, likelihood)
                    self.mean_module = gpytorch.means.ConstantMean()

                    # We only use the RBF kernel here, the WhiteNoiseKernel is deprecated
                    # and its featured integrated into the Likelihood-Module.
                    self.covar_module = gpytorch.kernels.ScaleKernel(gpytorch.kernels.RBFKernel())

                def forward(self, x):
                    mean_x = self.mean_module(x)
                    covar_x = self.covar_module(x)
                    return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)

            # initialize likelihood and model
            likelihood = gpytorch.likelihoods.GaussianLikelihood()
            model = ExactGPModel(train_x, train_y, likelihood)

            # Find optimal model hyperparameters
            model.train()
            likelihood.train()

            # Use the adam optimizer
            # Includes GaussianLikelihood parameters
            optimizer = torch.optim.Adam(model.parameters(), lr=lr)

            # "Loss" for GPs - the marginal log likelihood
            mll = gpytorch.mlls.ExactMarginalLogLikelihood(likelihood, model)

            for i in range(training_iter):
                # Zero gradients from previous iteration
                optimizer.zero_grad()
                # Output from model
                output = model(train_x)

                # Calc loss and backprop gradients
                loss = -mll(output, train_y)
                loss.backward()
                optimizer.step()

            # Get into evaluation (predictive posterior) mode
            model.eval()
            likelihood.eval()

            # Make predictions by feeding model through likelihood
            with torch.no_grad(), gpytorch.settings.fast_pred_var():
                mean = model(train_x).loc.detach()
                loglik = mll(model(train_x), train_y) * T

            resid = (train_y - mean).detach().numpy()
            mean = mean.detach().numpy()

        if return_means and not return_likelihood:
            return resid, mean
        elif return_likelihood and not return_means:
            return resid, loglik
        elif return_means and return_likelihood:
            return resid, mean, loglik
        return resid

    def _get_model_selection_criterion(self, j, parents, tau_max=0):
        """Returns log marginal likelihood for GP regression.

        Fits a GP model of the parents to variable j and returns the negative
        log marginal likelihood as a model selection score. Is used to determine
        optimal hyperparameters in PCMCI, in particular the pc_alpha value.

        Parameters
        ----------
        j : int
            Index of target variable in data array.

        parents : list
            List of form [(0, -1), (3, -2), ...] containing parents.

        tau_max : int, optional (default: 0)
            Maximum time lag. This may be used to make sure that estimates for
            different lags in X, Z, all have the same sample size.

        Returns:
        score : float
            Model score.
        """

        Y = [(j, 0)]
        X = [(j, 0)]   # dummy variable here
        Z = parents
        array, xyz, _ = \
            self.cond_ind_test.dataframe.construct_array(
                X=X, Y=Y, Z=Z,
                tau_max=tau_max,
                mask_type=self.cond_ind_test.mask_type,
                return_cleaned_xyz=False,
                do_checks=True,
                verbosity=self.verbosity)

        dim, T = array.shape

        _, logli = self._get_single_residuals(array,
                                              target_var=1,
                                              return_likelihood=True)

        score = -logli
        return score


class GPDCtorch(CondIndTest):
    r"""GPDC conditional independence test based on Gaussian processes and distance correlation. Here with gpytorch implementation.

    GPDC is based on a Gaussian process (GP) regression and a distance
    correlation test on the residuals [2]_. GP is estimated with gpytorch.
    The distance correlation test is implemented with the dcor package available
    from pip. Here the null distribution is not analytically available, but can be
    precomputed with the function generate_and_save_nulldists(...) which saves a
    \*.npz file containing the null distribution for different sample sizes.
    This file can then be supplied as null_dist_filename.

    Notes
    -----

    GPDC is based on a Gaussian process (GP) regression and a distance
    correlation test on the residuals. Distance correlation is described in
    [2]_. To test :math:`X \perp Y | Z`, first :math:`Z` is regressed out from
    :math:`X` and :math:`Y` assuming the  model

    .. math::  X & =  f_X(Z) + \epsilon_{X} \\
        Y & =  f_Y(Z) + \epsilon_{Y}  \\
        \epsilon_{X,Y} &\sim \mathcal{N}(0, \sigma^2)

    using GP regression. Here :math:`\sigma^2` and the kernel bandwidth are
    optimzed using ``gpytorch``. Then the residuals  are transformed to uniform
    marginals yielding :math:`r_X,r_Y` and their dependency is tested with

    .. math::  \mathcal{R}\left(r_X, r_Y\right)

    The null distribution of the distance correlation should be pre-computed.
    Otherwise it is computed during runtime.

    Parameters
    ----------
    null_dist_filename : str, otional (default: None)
        Path to file containing null distribution.

    **kwargs :
        Arguments passed on to parent class GaussProcRegTorch.

    """
    @property
    def measure(self):
        """
        Concrete property to return the measure of the independence test
        """
        return self._measure

    def __init__(self,
                 null_dist_filename=None,
                 **kwargs):
        self._measure = 'gp_dc'
        self.two_sided = False
        self.residual_based = True
        # Call the parent constructor
        CondIndTest.__init__(self, **kwargs)
        # Build the regressor
        self.gauss_pr = GaussProcRegTorch(self.sig_samples,
                                     self,
                                     null_dist_filename=null_dist_filename,
                                     verbosity=self.verbosity)

        if self.verbosity > 0:
            print("null_dist_filename = %s" % self.gauss_pr.null_dist_filename)
            print("")

    def _load_nulldist(self, filename):
        r"""
        Load a precomputed null distribution from a \*.npz file.  This
        distribution can be calculated using generate_and_save_nulldists(...).

        Parameters
        ----------
        filename : strng
            Path to the \*.npz file

        Returns
        -------
        null_dists, null_samples : dict, int
            The null distirbution as a dictionary of distributions keyed by
            sample size, the number of null samples in total.
        """
        return self.gauss_pr._load_nulldist(filename)

    def generate_nulldist(self, df, add_to_null_dists=True):
        """Generates null distribution for pairwise independence tests.

        Generates the null distribution for sample size df. Assumes pairwise
        samples transformed to uniform marginals. Uses get_dependence_measure
        available in class and generates self.sig_samples random samples. Adds
        the null distributions to self.gauss_pr.null_dists.

        Parameters
        ----------
        df : int
            Degrees of freedom / sample size to generate null distribution for.

        add_to_null_dists : bool, optional (default: True)
            Whether to add the null dist to the dictionary of null dists or
            just return it.

        Returns
        -------
        null_dist : array of shape [df,]
            Only returned,if add_to_null_dists is False.
        """
        return self.gauss_pr._generate_nulldist(df, add_to_null_dists)

    def generate_and_save_nulldists(self, sample_sizes, null_dist_filename):
        """Generates and saves null distribution for pairwise independence
        tests.

        Generates the null distribution for different sample sizes. Calls
        generate_nulldist. Null dists are saved to disk as
        self.null_dist_filename.npz. Also adds the null distributions to
        self.gauss_pr.null_dists.

        Parameters
        ----------
        sample_sizes : list
            List of sample sizes.

        null_dist_filename : str
            Name to save file containing null distributions.
        """
        self.gauss_pr._generate_and_save_nulldists(sample_sizes,
                                                   null_dist_filename)


    def _get_single_residuals(self, array, target_var,
                                    return_means=False,
                                    standardize=True,
                                    return_likelihood=False,
                                    training_iter=50,
                                    lr=0.1):
        """Returns residuals of Gaussian process regression.

        Performs a GP regression of the variable indexed by target_var on the
        conditions Z. Here array is assumed to contain X and Y as the first two
        rows with the remaining rows (if present) containing the conditions Z.
        Optionally returns the estimated mean and the likelihood.

        Parameters
        ----------
        array : array-like
            data array with X, Y, Z in rows and observations in columns

        target_var : {0, 1}
            Variable to regress out conditions from.

        standardize : bool, optional (default: True)
            Whether to standardize the array beforehand.

        return_means : bool, optional (default: False)
            Whether to return the estimated regression line.

        return_likelihood : bool, optional (default: False)
            Whether to return the log_marginal_likelihood of the fitted GP

        training_iter : int, optional (default: 50)
            Number of training iterations.

        lr : float, optional (default: 0.1)
            Learning rate (default: 0.1).

        Returns
        -------
        resid [, mean, likelihood] : array-like
            The residual of the regression and optionally the estimated mean
            and/or the likelihood.
        """
        return self.gauss_pr._get_single_residuals(
            array, target_var,
            return_means,
            standardize,
            return_likelihood,
            training_iter,
            lr)

    def get_model_selection_criterion(self, j, parents, tau_max=0):
        """Returns log marginal likelihood for GP regression.

        Fits a GP model of the parents to variable j and returns the negative
        log marginal likelihood as a model selection score. Is used to determine
        optimal hyperparameters in PCMCI, in particular the pc_alpha value.

        Parameters
        ----------
        j : int
            Index of target variable in data array.

        parents : list
            List of form [(0, -1), (3, -2), ...] containing parents.

        tau_max : int, optional (default: 0)
            Maximum time lag. This may be used to make sure that estimates for
            different lags in X, Z, all have the same sample size.

        Returns:
        score : float
            Model score.
        """
        return self.gauss_pr._get_model_selection_criterion(j, parents, tau_max)

    def get_dependence_measure(self, array, xyz, data_type=None):
        """Return GPDC measure.

        Estimated as the distance correlation of the residuals of a GP
        regression.

        Parameters
        ----------
        array : array-like
            data array with X, Y, Z in rows and observations in columns

        xyz : array of ints
            XYZ identifier array of shape (dim,).

        Returns
        -------
        val : float
            GPDC test statistic.
        """

        x_vals = self._get_single_residuals(array, target_var=0)
        y_vals = self._get_single_residuals(array, target_var=1)
        val = self._get_dcorr(np.array([x_vals, y_vals]))
        return val

    def _get_dcorr(self, array_resid):
        """Return distance correlation coefficient.

        The variables are transformed to uniform marginals using the empirical
        cumulative distribution function beforehand. Here the null distribution
        is not analytically available, but can be precomputed with the function
        generate_and_save_nulldists(...) which saves a *.npz file containing
        the null distribution for different sample sizes. This file can then be
        supplied as null_dist_filename.

        Parameters
        ----------
        array_resid : array-like
            data array must be of shape (2, T)

        Returns
        -------
        val : float
            Distance correlation coefficient.
        """
        # Remove ties before applying transformation to uniform marginals
        # array_resid = self._remove_ties(array_resid, verbosity=4)
        x_vals, y_vals = self._trafo2uniform(array_resid)
        val = dcor.distance_correlation(x_vals, y_vals, method='AVL')
        return val

    def get_shuffle_significance(self, array, xyz, value,
                                 return_null_dist=False,
                                 data_type=None):
        """Returns p-value for shuffle significance test.

        For residual-based test statistics only the residuals are shuffled.

        Parameters
        ----------
        array : array-like
            data array with X, Y, Z in rows and observations in columns

        xyz : array of ints
            XYZ identifier array of shape (dim,).

        value : number
            Value of test statistic for unshuffled estimate.

        Returns
        -------
        pval : float
            p-value
        """

        x_vals = self._get_single_residuals(array, target_var=0)
        y_vals = self._get_single_residuals(array, target_var=1)
        array_resid = np.array([x_vals, y_vals])
        xyz_resid = np.array([0, 1])

        null_dist = self._get_shuffle_dist(array_resid, xyz_resid,
                                           self.get_dependence_measure,
                                           sig_samples=self.sig_samples,
                                           sig_blocklength=self.sig_blocklength,
                                           verbosity=self.verbosity)

        # pval = (null_dist >= value).mean()
        pval = float(np.sum(null_dist >= value) + 1) / (self.sig_samples + 1)

        if return_null_dist:
            return pval, null_dist
        return pval

    def get_analytic_significance(self, value, T, dim, xyz):
        """Returns p-value for the distance correlation coefficient.

        The null distribution for necessary degrees of freedom (df) is loaded.
        If not available, the null distribution is generated with the function
        generate_nulldist(). It is recommended to generate the nulldists for a
        wide range of sample sizes beforehand with the function
        generate_and_save_nulldists(...). The distance correlation coefficient
        is one-sided. If the degrees of freedom are less than 1, numpy.nan is
        returned.

        Parameters
        ----------
        value : float
            Test statistic value.

        T : int
            Sample length

        dim : int
            Dimensionality, ie, number of features.

        xyz : array of ints
            XYZ identifier array of shape (dim,).

        Returns
        -------
        pval : float or numpy.nan
            p-value.
        """

        # GP regression approximately doesn't cost degrees of freedom
        df = T

        if df < 1:
            pval = np.nan
        else:
            # idx_near = (np.abs(self.sample_sizes - df)).argmin()
            if int(df) not in list(self.gauss_pr.null_dists):
                # if np.abs(self.sample_sizes[idx_near] - df) / float(df) > 0.01:
                if self.verbosity > 0:
                    print("Null distribution for GPDC not available "
                          "for deg. of freed. = %d." % df)
                self.generate_nulldist(df)
            null_dist_here = self.gauss_pr.null_dists[int(df)]
            pval = np.mean(null_dist_here > np.abs(value))
        return pval

